@@ -120,10 +120,57 @@ String ID3v2::Tag::comment() const
 
 String ID3v2::Tag::genre() const
 {
-  if(!d->frameListMap["TCON"].isEmpty()) {
-    String s = d->frameListMap["TCON"].front()->toString();
+  // TODO: In the next major version (TagLib 2.0) a list of multiple genres
+  // should be separated by " / " instead of " ".  For the moment to keep
+  // the behavior the same as released versions it is being left with " ".
 
-    // ID3v2 "content type" can contain a ID3v1 genre number in parenthesis at
+  if(!d->frameListMap["TCON"].isEmpty() &&
+     dynamic_cast<TextIdentificationFrame *>(d->frameListMap["TCON"].front()))
+  {
+    Frame *frame = d->frameListMap["TCON"].front();
+
+    // ID3v2.4 lists genres as the fields in its frames field list.  If the field
+    // is simply a number it can be assumed that it is an ID3v1 genre number.
+    // Here was assume that if an ID3v1 string is present that it should be
+    // appended to the genre string.  Multiple fields will be appended as the
+    // string is built.
+
+    if(d->header.majorVersion() == 4) {
+      TextIdentificationFrame *f = static_cast<TextIdentificationFrame *>(frame);
+      StringList fields = f->fieldList();
+
+      String genreString;
+      bool hasNumber = false;
+
+      for(StringList::ConstIterator it = fields.begin(); it != fields.end(); ++it) {
+        bool isNumber = true;
+        for(String::ConstIterator charIt = (*it).begin();
+            isNumber && charIt != (*it).end();
+            ++charIt)
+        {
+          isNumber = *charIt >= '0' && *charIt <= '9';
+        }
+
+        if(!genreString.isEmpty())
+          genreString.append(' ');
+
+        if(isNumber) {
+          int number = (*it).toInt();
+          if(number >= 0 && number <= 255) {
+            hasNumber = true;
+            genreString.append(ID3v1::genre(number));
+          }
+        }
+        else
+          genreString.append(*it);
+      }
+      if(hasNumber)
+        return genreString;
+    }
+
+    String s = frame->toString();
+
+    // ID3v2.3 "content type" can contain a ID3v1 genre number in parenthesis at
     // the beginning of the field.  If this is all that the field contains, do a
     // translation from that number to the name and return that.  If there is a
     // string folloing the ID3v1 genre number, that is considered to be
@@ -144,8 +191,8 @@ String ID3v2::Tag::genre() const
 
 TagLib::uint ID3v2::Tag::year() const
 {
-  if(!d->frameListMap["TRDC"].isEmpty())
-    return d->frameListMap["TRDC"].front()->toString().substr(0, 4).toInt();
+  if(!d->frameListMap["TDRC"].isEmpty())
+    return d->frameListMap["TDRC"].front()->toString().substr(0, 4).toInt();
   return 0;
 }
 
@@ -197,7 +244,7 @@ void ID3v2::Tag::setGenre(const String &s)
   int index = ID3v1::genreIndex(s);
 
   if(index != 255)
-    setTextFrame("TCON", "(" + String::number(index) + ")");
+    setTextFrame("TCON", String::number(index));
   else
     setTextFrame("TCON", s);
 }
@@ -205,10 +252,10 @@ void ID3v2::Tag::setGenre(const String &s)
 void ID3v2::Tag::setYear(uint i)
 {
   if(i <= 0) {
-    removeFrames("TRDC");
+    removeFrames("TDRC");
     return;
   }
-  setTextFrame("TRDC", String::number(i));
+  setTextFrame("TDRC", String::number(i));
 }
 
 void ID3v2::Tag::setTrack(uint i)
@@ -250,6 +297,11 @@ const FrameList &ID3v2::Tag::frameList() const
   return d->frameList;
 }
 
+const FrameList &ID3v2::Tag::frameList(const ByteVector &frameID) const
+{
+  return d->frameListMap[frameID];
+}
+
 void ID3v2::Tag::addFrame(Frame *frame)
 {
   d->frameList.append(frame);
@@ -336,7 +388,7 @@ void ID3v2::Tag::read()
 
 void ID3v2::Tag::parse(const ByteVector &data)
 {
-  uint frameDataOffset = 0;
+  uint frameDataPosition = 0;
   uint frameDataLength = data.size();
 
   // check for extended header
@@ -346,7 +398,7 @@ void ID3v2::Tag::parse(const ByteVector &data)
       d->extendedHeader = new ExtendedHeader;
     d->extendedHeader->setData(data);
     if(d->extendedHeader->size() <= data.size()) {
-      frameDataOffset += d->extendedHeader->size();
+      frameDataPosition += d->extendedHeader->size();
       frameDataLength -= d->extendedHeader->size();
     }
   }
@@ -360,18 +412,15 @@ void ID3v2::Tag::parse(const ByteVector &data)
 
   // parse frames
 
-  uint frameDataPosition = 0;
-
   // Make sure that there is at least enough room in the remaining frame data for
   // a frame header.
 
-  while(frameDataPosition < frameDataLength - Frame::headerSize()) {
+  while(frameDataPosition < frameDataLength - Frame::headerSize(d->header.majorVersion())) {
 
     // If the next data is position is 0, assume that we've hit the padding
     // portion of the frame data.
 
     if(data.at(frameDataPosition) == 0) {
-
       if(d->header.footerPresent())
         debug("Padding *and* a footer found.  This is not allowed by the spec.");
 
@@ -379,10 +428,8 @@ void ID3v2::Tag::parse(const ByteVector &data)
       return;
     }
 
-    bool synchSafeInts = d->header.majorVersion() >= 4;
-
-    Frame *frame = d->factory->createFrame(data.mid(frameDataOffset + frameDataPosition),
-                                           synchSafeInts);
+    Frame *frame = d->factory->createFrame(data.mid(frameDataPosition),
+                                           d->header.majorVersion());
 
     if(!frame)
       return;
@@ -394,7 +441,7 @@ void ID3v2::Tag::parse(const ByteVector &data)
       return;
     }
 
-    frameDataPosition += frame->size() + Frame::headerSize();
+    frameDataPosition += frame->size() + Frame::headerSize(d->header.majorVersion());
     addFrame(frame);
   }
 }
