@@ -1,5 +1,5 @@
 /***************************************************************************
-    copyright            : (C) 2002, 2003 by Scott Wheeler
+    copyright            : (C) 2002 - 2004 by Scott Wheeler
     email                : wheeler@kde.org
  ***************************************************************************/
 
@@ -26,6 +26,15 @@
 
 #include "tbytevector.h"
 
+// This is a bit ugly to keep writing over and over again.
+
+// A rather obscure feature of the C++ spec that I hadn't thought of that makes
+// working with C libs much more effecient.  There's more here:
+//
+// http://www.informit.com/isapi/product_id~{9C84DAB4-FE6E-49C5-BB0A-FB50331233EA}/content/index.asp
+
+#define DATA(x) (&(x->data[0]))
+
 namespace TagLib {
   static const uint crcTable[256] = {
     0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9, 0x130476dc, 0x17c56b6b,
@@ -180,6 +189,38 @@ namespace TagLib {
   private:
     const ByteVector v;
   };
+
+  template <class T>
+  T toNumber(const std::vector<char> &data, bool mostSignificantByteFirst)
+  {
+    T sum = 0;
+
+    if(data.size() <= 0) {
+      debug("ByteVectorMirror::toNumber<T>() -- data is empty, returning 0");
+      return sum;
+    }
+
+    uint size = sizeof(T);
+    uint last = data.size() > size ? size - 1 : data.size() - 1;
+
+    for(uint i = 0; i <= last; i++)
+      sum |= (T) uchar(data[i]) << ((mostSignificantByteFirst ? last - i : i) * 8);
+
+    return sum;
+  }
+
+  template <class T>
+  ByteVector fromNumber(T value, bool mostSignificantByteFirst)
+  {
+    int size = sizeof(T);
+
+    ByteVector v(size, 0);
+
+    for(int i = 0; i < size; i++)
+      v[i] = uchar(value >> ((mostSignificantByteFirst ? size - 1 - i : i) * 8) & 0xff);
+
+    return v;
+  }
 }
 
 using namespace TagLib;
@@ -218,22 +259,17 @@ ByteVector ByteVector::fromCString(const char *s, uint length)
 
 ByteVector ByteVector::fromUInt(uint value, bool mostSignificantByteFirst)
 {
-  ByteVector v(4, 0);
+  return fromNumber<uint>(value, mostSignificantByteFirst);
+}
 
-  for(int i = 0; i < 4; i++)
-    v[i] = uchar(value >> ((mostSignificantByteFirst ? 3 - i : i) * 8) & 0xff);
-
-  return v;
+ByteVector ByteVector::fromShort(short value, bool mostSignificantByteFirst)
+{
+  return fromNumber<short>(value, mostSignificantByteFirst);
 }
 
 ByteVector ByteVector::fromLongLong(long long value, bool mostSignificantByteFirst)
 {
-  ByteVector v(8, 0);
-
-  for(int i = 0; i < 8; i++)
-    v[i] = uchar(value >> ((mostSignificantByteFirst ? 7 - i : i) * 8) & 0xff);
-
-  return v;
+  return fromNumber<long long>(value, mostSignificantByteFirst);
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -285,7 +321,7 @@ void ByteVector::setData(const char *data, uint length)
   detach();
 
   resize(length);
-  ::memcpy(&(d->data[0]), data, length);
+  ::memcpy(DATA(d), data, length);
 }
 
 void ByteVector::setData(const char *data)
@@ -295,18 +331,13 @@ void ByteVector::setData(const char *data)
 
 char *ByteVector::data()
 {
-  // A rather obscure feature of the C++ spec that I hadn't thought of that makes
-  // working with C libs much more effecient.  There's more here:
-  //
-  // http://www.informit.com/isapi/product_id~{9C84DAB4-FE6E-49C5-BB0A-FB50331233EA}/content/index.asp
-
   detach();
-  return &(d->data[0]);
+  return DATA(d);
 }
 
 const char *ByteVector::data() const
 {
-  return &(d->data[0]);
+  return DATA(d);
 }
 
 ByteVector ByteVector::mid(uint index, uint length) const
@@ -402,7 +433,7 @@ void ByteVector::append(const ByteVector &v)
 
   uint originalSize = d->size;
   resize(d->size + v.d->size);
-  ::memcpy(&(d->data[0]) + originalSize, v.data(), v.size());
+  ::memcpy(DATA(d) + originalSize, DATA(v.d), v.size());
 }
 
 void ByteVector::clear()
@@ -418,15 +449,15 @@ TagLib::uint ByteVector::size() const
 
 ByteVector &ByteVector::resize(uint size, char padding)
 {
-  d->size = size;
-
-  if(d->data.size() < size) {
+  if(d->size < size) {
     d->data.reserve(size);
-    d->data.insert(d->data.end(), size - d->data.size(), padding);
+    d->data.insert(d->data.end(), size - d->size, padding);
   }
   else
     d->data.erase(d->data.begin() + size, d->data.end());
 
+  d->size = size;
+
   return *this;
 }
 
@@ -470,27 +501,17 @@ TagLib::uint ByteVector::checksum() const
 
 TagLib::uint ByteVector::toUInt(bool mostSignificantByteFirst) const
 {
-  uint sum = 0;
-  int last = d->data.size() > 4 ? 3 : d->data.size() - 1;
+  return toNumber<uint>(d->data, mostSignificantByteFirst);
+}
 
-  for(int i = 0; i <= last; i++)
-    sum |= uchar(d->data[i]) << ((mostSignificantByteFirst ? last - i : i) * 8);
-
-  return sum;
+short ByteVector::toShort(bool mostSignificantByteFirst) const
+{
+  return toNumber<unsigned short>(d->data, mostSignificantByteFirst);
 }
 
 long long ByteVector::toLongLong(bool mostSignificantByteFirst) const
 {
-  // Just do all of the bit operations on the unsigned value and use an implicit
-  // cast on the way out.
-
-  unsigned long long sum = 0;
-  int last = d->data.size() > 8 ? 7 : d->data.size() - 1;
-
-  for(int i = 0; i <= last; i++)
-    sum |= (unsigned long long) uchar(d->data[i]) << ((mostSignificantByteFirst ? last - i : i) * 8);
-
-  return sum;
+  return toNumber<unsigned long long>(d->data, mostSignificantByteFirst);
 }
 
 const char &ByteVector::operator[](int index) const
@@ -507,15 +528,10 @@ char &ByteVector::operator[](int index)
 
 bool ByteVector::operator==(const ByteVector &v) const
 {
-  if(size() != v.size())
+  if(d->size != v.d->size)
     return false;
 
-  for(uint i = 0; i < size(); i++) {
-    if(at(i) != v.at(i))
-      return false;
-  }
-
-  return true;
+  return ::memcmp(data(), v.data(), size()) == 0;
 }
 
 bool ByteVector::operator!=(const ByteVector &v) const
@@ -525,7 +541,10 @@ bool ByteVector::operator!=(const ByteVector &v) const
 
 bool ByteVector::operator==(const char *s) const
 {
-  return operator==(fromCString(s));
+  if(d->size != ::strlen(s))
+    return false;
+
+  return ::memcmp(data(), s, d->size) == 0;
 }
 
 bool ByteVector::operator!=(const char *s) const
